import unittest

from bamboo import ASGIHTTPApp, ASGIHTTPEndpoint
from bamboo.request import http

from . import get_log_name
from .asgi_util import ASGIHTTPServerForm, ASGIHTTPTestExecutor


app = ASGIHTTPApp()
PATH_SERVER_LOG = get_log_name(__file__)


@app.route("hello")
class MockEndpoint(ASGIHTTPEndpoint):

    async def do_GET(self) -> None:
        self.send_body(b"Hello, World!")


@app.route("attributes")
class AttributesEndpoint(ASGIHTTPEndpoint):

    async def do_GET(self) -> None:
        assert self.asgi_version == "3.0"
        assert self.http_version == "1.1"
        assert self.scheme == "http"
        assert self.get_host_addr() == ("localhost", 8000)
        assert self.path == "/attributes"
        assert self.queries == {}
        assert self.method == "GET"

        self.send_only_status()


class ASGIAppTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form = ASGIHTTPServerForm("localhost", 8000, app, PATH_SERVER_LOG)
        cls.executor = ASGIHTTPTestExecutor(form).start_serve()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor.close()

    def test_hello(self):
        with http.get("http://localhost:8000/hello") as res:
            self.assertEqual(res.body, b"Hello, World!")

    def test_attributes(self):
        with http.get("http://localhost:8000/attributes") as res:
            self.assertTrue(res.ok)


if __name__ == "__main__":
    unittest.main()



import unittest

from bamboo import (
    ContentType,
    MediaTypes,
    XWWWFormUrlEncodedData,
    ApiValidationFailedError,
)


ideal_content_type = ContentType(MediaTypes.x_www_form_urlencoded, "utf-8")


class TestData(XWWWFormUrlEncodedData):

    name: str
    age: str
    email: str


data_raw = b"name=hogehoge&email=hoge@hoge.com&age=18"
data_key_not_included = b"name=hogehoge&age=20"
data_duplicated_key = b"name=hogehoge&name=hogest&age=20&email=hoge@hoge.com"


class TestXWWWFromUrlEncodedData(unittest.TestCase):

    def test_data_raw(self):
        data = TestData.__validate__(data_raw, ideal_content_type)
        self.assertEqual(data.name, "hogehoge")
        self.assertEqual(int(data.age), 18)
        self.assertEqual(data.email, "hoge@hoge.com")

    def test_data_key_not_included(self):
        with self.assertRaises(ApiValidationFailedError) as err:
            data = TestData.__validate__(data_key_not_included, ideal_content_type)
        self.assertIsInstance(err.exception, ApiValidationFailedError)

    def test_data_duplicated_key(self):
        with self.assertRaises(ApiValidationFailedError) as err:
            data = TestData.__validate__(data_duplicated_key, ideal_content_type)
        self.assertIsInstance(err.exception, ApiValidationFailedError)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import dataclasses
import multiprocessing
import os
import signal
import sys
import time
import typing as t

from bamboo import (
    ASGIHTTPApp,
)
from bamboo.util.string import ColorCode, insert_colorcode
import uvicorn


__all__ = [
    "ASGIHTTPServerForm",
    "ASGIHTTPTestExecutor",
]


@dataclasses.dataclass
class ASGIHTTPServerForm:

    host: str
    port: int
    app: ASGIHTTPApp
    path_log: str


def serve_at(form: ASGIHTTPServerForm) -> None:
    f_log = open(form.path_log, "wt")
    sys.stdout = f_log
    sys.stdin = f_log

    if not len(form.host):
        form.host = "localhost"
    uvicorn.run(
        form.app,
        host=form.host,
        port=form.port,
        log_level="debug"
    )


class ASGIHTTPTestExecutor:

    def __init__(self, *forms: ASGIHTTPServerForm) -> None:
        self._forms: t.List[ASGIHTTPServerForm] = []
        self._children: t.List[multiprocessing.Process] = []

        self.add_forms(*forms)

    def add_forms(self, *forms: ASGIHTTPServerForm) -> None:
        for form in forms:
            self._forms.append(form)

    def start_serve(self, waiting: float = 0.1) -> ASGIHTTPTestExecutor:
        for form in self._forms:
            child = multiprocessing.Process(target=serve_at, args=(form,))
            child.start()
            self._children.append(child)
        time.sleep(waiting)

        return self

    def close(self, pop: bool = True) -> None:
        for child in self._children:
            os.kill(child.pid, signal.SIGINT)
            child.join()
            child.close()

        if pop:
            self._children.clear()

    def exec(
        self,
        func: t.Callable[[t.Tuple[t.Any, ...]], None],
        args: t.Tuple[t.Any, ...] = (),
        waiting: float = 0.05,
    ) -> None:
        with self.start_serve(waiting=waiting):
            func(*args)

    @classmethod
    def debug(
        cls,
        app: ASGIHTTPApp,
        path_log: str,
        host: str = "localhost",
        port: int = 8000,
        waiting: float = 0.05
    ) -> None:
        form = ASGIHTTPServerForm(host, port, app, path_log)
        executor = cls(form)

        try:
            print(f"Hosting on {host}:{port} ...")
            print(insert_colorcode(
                "WARNING: This is debug mode. "
                "Do not use it in your production deployment.",
                ColorCode.RED
            ))

            executor.start_serve(waiting=waiting)
            while True:
                time.sleep(60 * 60)
        finally:
            executor.close()
            print(
                "\nHosting terminated. The log was output into " +
                insert_colorcode(f"{path_log}", ColorCode.YELLOW) + "."
            )

    def __enter__(self) -> None:
        pass

    def __exit__(self, type, value, traceback) -> None:
        self.close()



import unittest

from bamboo.util.string import CircularChar, SerialString


class TestUtilString(unittest.TestCase):

    def test_char(self):
        char = CircularChar()
        for _ in range(char.END):
            char.next()

        self.assertEqual(char.current, CircularChar.TAIL)

    def test_serial_string_up(self):
        serial = SerialString(2)

        for _ in range(CircularChar.END + 1):
            for _ in range(CircularChar.END + 1):
                serial.next()
            self.assertEqual(serial.current[-1], CircularChar.TAIL)

        full = serial.next()
        self.assertTrue(full is None)

    def test_serial_string_nonup(self):
        serial = SerialString(2)

        self.assertEqual(serial.next(up=False), "00")
        self.assertEqual(serial.current, "00")


if __name__ == "__main__":
    unittest.main()



import typing as t


def make_header(name: str, value: str, **params: str) -> t.Tuple[str, str]:
    """Make pair of header field and its value with other directives.

    Args:
        name: Field name of the header.
        value: Value of the field.
        **params: Directives added to the field.
    """
    params = [f"; {header}={val}" for header, val in params.items()]
    params = "".join(params)
    return (name, value + params)



from __future__ import annotations
import typing as t


__all__ = [
    "awaitable_property",
    "awaitable_cached_property",
    "cached_property",
    "class_property",
    "joint_funcs",
    "joint_methods",
]


def joint_funcs(*func: t.Callable[[], None]) -> t.Callable[[], None]:
    def joint():
        for f in func:
            f()
    return joint


def joint_methods(
    *methods: t.Callable[[t.Any], None],
) -> t.Callable[[t.Any], None]:
    def joint(obj: t.Any):
        for m in methods:
            m(obj)
    return joint


Object = t.TypeVar("Object")
ReturnGetter = t.TypeVar("ReturnGetter")


class cached_property(t.Generic[Object, ReturnGetter]):

    def __init__(
        self,
        fget: t.Callable[[Object], ReturnGetter]
    ) -> None:
        self.__doc__ = getattr(fget, "__doc__")
        self._fget = fget
        self._obj2val: t.Dict[Object, ReturnGetter] = {}

    def __get__(
        self,
        obj: t.Union[Object, None],
        clazz: t.Optional[t.Type[Object]] = None
    ) -> t.Union[cached_property, ReturnGetter]:
        if obj is None:
            return self
        if self._fget is not None:
            val = self._obj2val.get(obj)
            if val is None:
                val = self._fget(obj)
                self._obj2val[obj] = val
            return val
        raise AttributeError("'getter' has not been set yet.")

    def getter(
        self,
        fget: t.Callable[[Object], ReturnGetter]
    ) -> cached_property:
        self._fget = fget
        return self


class class_property(t.Generic[Object, ReturnGetter]):

    def __init__(
        self,
        fget: t.Callable[[Object], ReturnGetter]
    ) -> None:
        self.__doc__ = getattr(fget, "__doc__")
        self._fget = fget

    def __get__(
        self,
        obj: t.Union[Object, None],
        clazz: t.Optional[t.Type[Object]] = None
    ) -> ReturnGetter:
        if clazz is None:
            clazz = type(obj)

        if self._fget is not None:
            return self._fget(clazz)
        raise AttributeError("'getter' has not been set yet.")

    def getter(
        self,
        fget: t.Callable[[Object], ReturnGetter]
    ) -> class_property:
        self._fget = fget
        return self


class awaitable_property(t.Generic[Object, ReturnGetter]):

    def __init__(
        self,
        fget: t.Callable[[Object], t.Awaitable[ReturnGetter]]
    ) -> None:
        self.__doc__ = getattr(fget, "__doc__")
        self._fget = fget

    async def __get__(
        self,
        obj: t.Union[Object, None],
        clazz: t.Optional[t.Type[Object]] = None
    ) -> t.Union[awaitable_property, ReturnGetter]:
        if obj is None:
            return self
        if self._fget is not None:
            return await self._fget(obj)
        raise AttributeError("'getter' has not been set yet.")

    def getter(
        self,
        fget: t.Callable[[Object], t.Awaitable[ReturnGetter]]
    ) -> awaitable_property:
        self._fget = fget
        return self


class awaitable_cached_property(t.Generic[Object, ReturnGetter]):

    def __init__(
        self,
        fget: t.Callable[[Object], t.Awaitable[ReturnGetter]]
    ) -> None:
        self.__doc__ = getattr(fget, "__doc__")
        self._fget = fget
        self._obj2val: t.Dict[Object, ReturnGetter] = {}

    async def __get__(
        self,
        obj: t.Union[Object, None],
        clazz: t.Optional[t.Type[Object]] = None
    ) -> t.Union[awaitable_property, ReturnGetter]:
        if obj is None:
            return self
        if self._fget is not None:
            val = self._obj2val.get(obj)
            if val is None:
                val = await self._fget(obj)
                self._obj2val[obj] = val
            return val
        raise AttributeError("'getter' has not been set yet.")

    def getter(
        self,
        fget: t.Callable[[Object], t.Awaitable[ReturnGetter]]
    ) -> awaitable_cached_property:
        self._fget = fget
        return self



import unittest

from bamboo import (
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http

from . import get_log_name


PATH_SERVER_LOG = get_log_name(__file__)
sub_app1 = WSGIApp()
sub_app2 = WSGIApp()
app = WSGIApp()


@sub_app1.route("client", version=(1, 2, 3))
class ClientEndpoint(WSGIEndpoint):

    def do_GET(self) -> None:
        self.send_body(b"Hello, Client!")


@sub_app2.route("hoge")
class HogeEndpoint(WSGIEndpoint):

    def do_GET(self) -> None:
        self.send_body(b"hoge")


@app.route("world")
class WorldEndpoint(WSGIEndpoint):

    def do_GET(self) -> None:
        self.send_body(b"Hello, World!")


app.graft(sub_app1)
app.graft(sub_app2, onto=("hoge",))


class TestGrafting(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form = WSGIServerForm("", 8000, app, PATH_SERVER_LOG)
        cls.executor = WSGITestExecutor(form).start_serve()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor.close()

    def test_world(self) -> None:
        with http.get("http://localhost:8000/world") as res:
            self.assertTrue(res.ok)

    def test_client(self) -> None:
        for i in range(1, 4):
            with http.get(f"http://localhost:8000/v{i}/client") as res:
                self.assertTrue(res.ok)

    def test_hoge(self) -> None:
        with http.get("http://localhost:8000/hoge/hoge") as res:
            self.assertTrue(res.ok)


if __name__ == "__main__":
    unittest.main()



import unittest

from bamboo import (
    ASGIHTTPApp,
    ASGIHTTPEndpoint,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http
from bamboo.sticky.http import ClientInfo, restricts_client

from . import get_log_name
from .asgi_util import ASGIHTTPServerForm, ASGIHTTPTestExecutor


app_asgi = ASGIHTTPApp()
app_wsgi = WSGIApp()
PATH_ASGI_SERVER_LOG = get_log_name(__file__, "asgi")
PATH_WSGI_SERVER_LOG = get_log_name(__file__, "wsgi")
CLIENT_CORRECT = ClientInfo("127.0.0.1")
CLIENT_INCORRECT = ClientInfo("199.9.9.9")


@app_asgi.route()
class TestASGIHTTPEndpoint(ASGIHTTPEndpoint):

    @restricts_client(CLIENT_CORRECT)
    async def do_GET(self) -> None:
        self.send_only_status()

    @restricts_client(CLIENT_INCORRECT)
    async def do_HEAD(self) -> None:
        self.send_only_status()


@app_wsgi.route()
class TestWSGIEndpoint(WSGIEndpoint):

    @restricts_client(CLIENT_CORRECT)
    def do_GET(self) -> None:
        self.send_only_status()

    @restricts_client(CLIENT_INCORRECT)
    def do_HEAD(self) -> None:
        self.send_only_status()


class TestStickyRestrictsClient(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form_asgi = ASGIHTTPServerForm("", 8000, app_asgi, PATH_ASGI_SERVER_LOG)
        form_wsgi = WSGIServerForm("", 8001, app_wsgi, PATH_WSGI_SERVER_LOG)
        cls.executor_asgi = ASGIHTTPTestExecutor(form_asgi).start_serve()
        cls.executor_wsgi = WSGITestExecutor(form_wsgi).start_serve()
        cls.uri_asgi = "http://localhost:8000"
        cls.uri_wsgi = "http://localhost:8001"

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor_asgi.close()
        cls.executor_wsgi.close()

    def test_asgi_correct(self):
        with http.get(self.uri_asgi) as res:
            self.assertTrue(res.ok)

    def test_asgi_incorrect(self):
        with http.head(self.uri_asgi) as res:
            self.assertFalse(res.ok)

    def test_wsgi_correct(self):
        with http.get(self.uri_wsgi) as res:
            self.assertTrue(res.ok)

    def test_wsgi_incorrect(self):
        with http.head(self.uri_wsgi) as res:
            self.assertFalse(res.ok)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import http.client
import typing as t

from bamboo.api import BinaryApiData
from bamboo.base import ContentType
from bamboo.request import ResponseData_t
from bamboo.util.deco import cached_property


__all__ = []


class ResponseBodyAlreadyReadError(Exception):
    """Response body has already been read and data consistensy would be
    broken.
    """
    pass


class Response(t.Generic[ResponseData_t]):
    """Response returned by request functions of `http` or `https` modules.

    Examples:
        ```python
        from bamboo.request import http

        with http.get("http://localhost:8000/hello") as res:
            headers = res.headers

            if res.ok:
                body = res.body
                print(body)
            else:
                print(f"Error occured. Response status: {res.status}")

            # The session will be automatically closed by leaving the block.
        ```
    """

    def __init__(
        self,
        conn: http.client.HTTPConnection,
        res: http.client.HTTPResponse,
        datacls: t.Type[ResponseData_t] = BinaryApiData,
    ) -> None:
        """
        Args:
            conn: connection object of a session.
            res: HTTPResponse of a request.
            datacls: ApiData class to attach raw response body.
        """
        self._conn = conn
        self._res = res
        self._datacls = datacls
        self._is_read = False

    @property
    def headers(self) -> http.client.HTTPMessage:
        """All response headers.
        """
        return self._res.msg

    def get_header(self, name: str) -> t.Optional[str]:
        """Retrive header value from response headers.

        Args:
            name: Header name.

        Returns:
            Value of header if existing, None otherwise.
        """
        return self._res.getheader(name)

    @property
    def url(self) -> str:
        """Real url of the endpoint.
        """
        return self._res.geturl()

    @property
    def status(self) -> int:
        """Response status.
        """
        return self._res.status

    @property
    def version(self) -> int:
        """HTTP version of the session.
        """
        return self._res.version

    @property
    def ok(self) -> bool:
        """If request succeeded or not.
        """
        status = self.status
        if 200 <= status < 300:
            return True
        return False

    @property
    def is_closed(self) -> bool:
        """If the session is closed or not.
        """
        return self._res.closed

    @property
    def fileno(self) -> int:
        """File number of the socket.
        """
        return self._res.fileno()

    @property
    def content_length(self) -> t.Optional[int]:
        """Content length of the response if existing, None otherwise.
        """
        length = self.get_header("Content-Length")
        if length is not None:
            return int(length)
        return None

    def read(self, amt: t.Optional[int] = None) -> bytes:
        """Reads and returns the response body.

        Args:
            amt: Amount of the binary, all of it if None.

        Returns:
            Results of reading. Its length may be less than `amt`.
        """
        if not self._is_read:
            self._is_read = True
        return self._res.read(amt)

    @cached_property
    def body(self) -> bytes:
        """The raw response body.

        This property returns all reponse body and caches the result.
        If you want to read the response body step-by-step by chunks,
        then you can use the `read` method, but cannot use both the
        property and the `read` method.

        Raises:
            ResponseBodyAlreadyReadError: Raised if the `read` method has
                been already used.
        """
        if self._is_read:
            raise ResponseBodyAlreadyReadError(
                "Response body has been already read by 'read' method. "
                "Data consistensy would be broken."
            )
        return self._res.read(self.content_length)

    def attach(
        self,
        datacls: t.Optional[t.Type[ResponseData_t]] = None,
    ) -> ResponseData_t:
        """Generate new object of specified ApiData from response body.

        Args:
            datacls: ApiData class or its subclass.

        Returns:
            ResponseData_t: Generated object.
        """
        content_type_raw = self.get_header("Content-Type")
        if content_type_raw:
            content_type = ContentType.parse(content_type_raw)

        if datacls is None:
            if content_type_raw is None:
                content_type = self._datacls.__content_type__
            return self._datacls.__validate__(self.body, content_type)
        else:
            if content_type_raw is None:
                content_type = datacls.__content_type__
            return datacls.__validate__(self.body, content_type)

    def close(self) -> None:
        """Close the session.
        """
        self._conn.close()

    def __enter__(self) -> Response[ResponseData_t]:
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()



import unittest

from bamboo import (
    ASGIHTTPApp,
    ASGIHTTPEndpoint,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http
from bamboo.sticky.http import has_query_of
from bamboo.util.string import rand_string

from . import get_log_name
from .asgi_util import ASGIHTTPServerForm, ASGIHTTPTestExecutor


app_asgi = ASGIHTTPApp()
app_wsgi = WSGIApp()
PATH_ASGI_SERVER_LOG = get_log_name(__file__, "asgi")
PATH_WSGI_SERVER_LOG = get_log_name(__file__, "wsgi")
RANDOM_QUERIES = [(rand_string(10), rand_string(10)) for _ in range(10)]


@app_asgi.route()
class TestASGIHTTPEndpoint(ASGIHTTPEndpoint):

    @has_query_of(RANDOM_QUERIES[0][0])
    @has_query_of(RANDOM_QUERIES[1][0])
    @has_query_of(RANDOM_QUERIES[2][0])
    @has_query_of(RANDOM_QUERIES[3][0])
    @has_query_of(RANDOM_QUERIES[4][0])
    @has_query_of(RANDOM_QUERIES[5][0])
    @has_query_of(RANDOM_QUERIES[6][0])
    @has_query_of(RANDOM_QUERIES[7][0])
    @has_query_of(RANDOM_QUERIES[8][0])
    @has_query_of(RANDOM_QUERIES[9][0])
    async def do_GET(
        self,
        header0: str,
        header1: str,
        header2: str,
        header3: str,
        header4: str,
        header5: str,
        header6: str,
        header7: str,
        header8: str,
        header9: str,
    ) -> None:
        self.send_json({
            "headers": [
                header0,
                header1,
                header2,
                header3,
                header4,
                header5,
                header6,
                header7,
                header8,
                header9,
            ]
        })


@app_wsgi.route()
class TestWSGIEndpoint(WSGIEndpoint):

    @has_query_of(RANDOM_QUERIES[0][0])
    @has_query_of(RANDOM_QUERIES[1][0])
    @has_query_of(RANDOM_QUERIES[2][0])
    @has_query_of(RANDOM_QUERIES[3][0])
    @has_query_of(RANDOM_QUERIES[4][0])
    @has_query_of(RANDOM_QUERIES[5][0])
    @has_query_of(RANDOM_QUERIES[6][0])
    @has_query_of(RANDOM_QUERIES[7][0])
    @has_query_of(RANDOM_QUERIES[8][0])
    @has_query_of(RANDOM_QUERIES[9][0])
    def do_GET(
        self,
        header0: str,
        header1: str,
        header2: str,
        header3: str,
        header4: str,
        header5: str,
        header6: str,
        header7: str,
        header8: str,
        header9: str,
    ) -> None:
        self.send_json({
            "headers": [
                header0,
                header1,
                header2,
                header3,
                header4,
                header5,
                header6,
                header7,
                header8,
                header9,
            ]
        })


class TestStickyHasQueryOf(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form_asgi = ASGIHTTPServerForm("", 8000, app_asgi, PATH_ASGI_SERVER_LOG)
        form_wsgi = WSGIServerForm("", 8001, app_wsgi, PATH_WSGI_SERVER_LOG)
        cls.executor_asgi = ASGIHTTPTestExecutor(form_asgi).start_serve()
        cls.executor_wsgi = WSGITestExecutor(form_wsgi).start_serve()
        cls.uri_asgi = "http://localhost:8000"
        cls.uri_wsgi = "http://localhost:8001"

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor_asgi.close()
        cls.executor_wsgi.close()

    def test_asgi(self) -> None:
        with http.get(
            self.uri_asgi,
            query={name: [value] for name, value in RANDOM_QUERIES},
        ) as res:
            self.assertTrue(res.ok)

    def test_wsgi(self) -> None:
        with http.get(
            self.uri_wsgi,
            query={name: [value] for name, value in RANDOM_QUERIES},
        ) as res:
            self.assertTrue(res.ok)


if __name__ == "__main__":
    unittest.main()
